﻿/*
 * Copyright 2020 faddenSoft
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
using System;
using System.Collections.ObjectModel;

using PluginCommon;

namespace Budge3d {
    /// <summary>
    /// Visualizer for Budge 3D meshes.
    ///
    /// The code rotates about Z, then Y, then X, but the coordinate system is weird, so it's
    /// effectively Y-X-Z, which in matrix form is R = Rz * Rx * Ry.
    /// </summary>
    public class Budge3dMesh : MarshalByRefObject, IPlugin, IPlugin_Visualizer_v2 {
        // IPlugin
        public string Identifier {
            get { return "Budge3D Mesh Visualizer"; }
        }
        private IApplication mAppRef;
        private byte[] mFileData;

        // Visualization identifiers; DO NOT change or projects that use them will break.
        private const string VIS_GEN_BUDGE3D_MESH = "budge3d-mesh";

        private const string P_OFFSET = "offset";
        private const string P_SHAPE_INDEX = "shapeIndex";

        // Visualization descriptors.
        private VisDescr[] mDescriptors = new VisDescr[] {
            new VisDescr(VIS_GEN_BUDGE3D_MESH, "Budge3D Mesh", VisDescr.VisType.Wireframe,
                new VisParamDescr[] {
                    new VisParamDescr("File offset (hex)",
                        P_OFFSET, typeof(int), 0, 0x00ffffff, VisParamDescr.SpecialMode.Offset, 0),
                    new VisParamDescr("Shape index",
                        P_SHAPE_INDEX, typeof(int), 0, 15, 0, 0),

                    // This is interpreted by the main app.  The default is "true", so we need
                    // to include it so we can make the default "false".
                    VisWireframe.Param_IsPerspective("Perspective projection", false),
                }),
        };


        // IPlugin
        public void Prepare(IApplication appRef, byte[] fileData, AddressTranslate addrTrans) {
            mAppRef = appRef;
            mFileData = fileData;
        }

        // IPlugin
        public void Unprepare() {
            mAppRef = null;
            mFileData = null;
        }

        // IPlugin_Visualizer
        public VisDescr[] GetVisGenDescrs() {
            if (mFileData == null) {
                return null;
            }
            return mDescriptors;
        }

        // IPlugin_Visualizer
        public IVisualization2d Generate2d(VisDescr descr,
                ReadOnlyDictionary<string, object> parms) {
            mAppRef.ReportError("2d not supported");
            return null;
        }

        // IPlugin_Visualizer_v2
        public IVisualizationWireframe GenerateWireframe(VisDescr descr,
                ReadOnlyDictionary<string, object> parms) {
            switch (descr.Ident) {
                case VIS_GEN_BUDGE3D_MESH:
                    return GenerateWireframe(parms);
                default:
                    mAppRef.ReportError("Unknown ident " + descr.Ident);
                    return null;
            }
        }

        private IVisualizationWireframe GenerateWireframe(ReadOnlyDictionary<string, object> parms) {
            int offset = Util.GetFromObjDict(parms, P_OFFSET, 0);
            int shapeIndex = Util.GetFromObjDict(parms, P_SHAPE_INDEX, 0);

            if (offset < 0 || offset >= mFileData.Length) {
                // should be caught by editor
                mAppRef.ReportError("Invalid parameter");
                return null;
            }

            // Offset should point at NumObjects.
            int xcOffset = offset + (0x800 - 0x7fd);
            int ycOffset = offset + (0x900 - 0x7fd);
            int zcOffset = offset + (0xa00 - 0x7fd);
            int edgeStartOffset = offset + (0xb00 - 0x7fd);
            int edgeEndOffset = offset + (0xc00 - 0x7fd);
            int firstVertexOffset = offset + (0xd00 - 0x7fd);
            int lastVertexOffset = offset + (0xd10 - 0x7fd);
            int firstEdgeOffset = offset + (0xd20 - 0x7fd);
            int lastEdgeOffset = offset + (0xd30 - 0x7fd);

            VisWireframe vw = new VisWireframe();
            try {
                int numObjects = mFileData[offset];
                if (shapeIndex < 0 || shapeIndex >= numObjects) {
                    mAppRef.ReportError("Invalid shape index (max " + numObjects + ")");
                    return null;
                }

                int firstVertex = mFileData[firstVertexOffset + shapeIndex];
                int lastVertex = mFileData[lastVertexOffset + shapeIndex] - 1;
                int firstEdge = mFileData[firstEdgeOffset + shapeIndex];
                int lastEdge = mFileData[lastEdgeOffset + shapeIndex] - 1;

                mAppRef.DebugLog("Shape " + shapeIndex + ": vert " + firstVertex + "-" +
                    lastVertex + ", edge " + firstEdge + "-" + lastEdge);

                // The coordinate system shown in the manual for the space shuttle shape
                // puts the nose toward -X, extends the wings toward +/- Y, and extends the tail
                // toward +Z.  We switch the axes around to get a more conventional
                // arrangement.
                for (int vertex = firstVertex; vertex <= lastVertex; vertex++) {
                    vw.AddVertex((sbyte)(mFileData[ycOffset + vertex]),
                        (sbyte)(mFileData[zcOffset + vertex]),
                        (sbyte)(mFileData[xcOffset + vertex]));
                }
                for (int edge = firstEdge; edge <= lastEdge; edge++) {
                    vw.AddEdge(mFileData[edgeStartOffset + edge] - firstVertex,
                        mFileData[edgeEndOffset + edge] - firstVertex);
                }
            } catch (IndexOutOfRangeException) {
                // assume it was our file data access that caused the failure
                mAppRef.ReportError("Ran off end of file");
                return null;
            }

            string msg;
            if (!vw.Validate(out msg)) {
                mAppRef.ReportError("Data error: " + msg);
                return null;
            }

            return vw;
        }
    }
}
